"""This module hosts all method for loading/saving new models.

All methods for loading / saving a new model are within this module. As well as
the functionality to get/set the currently loaded model. Whenever a function needs
the currently loaded model, and no model was provided in the `kwargs`, the
:func:`.get_current_model()` function retrieves the model loaded last. There are also convenience
methods to load models directly from BioModels or from url.

"""
import COPASI
import os
try:
    import urllib2
    _use_urllib2 = True
except ImportError:
    import urllib
    import urllib.request
    _use_urllib2 = False

import traceback
import logging
import glob
import atexit
import subprocess
import tempfile
import shutil
import platform

__current_model = None
__model_list = []
__temp_files = []
__temp_dirs = []


def set_current_model(model):
    """Sets the current model.

    The current model, is the one that all functions will use if not is provided explicitly.

    :param model: the model to be set as current
    :type model: COPASI.CDataModel
    :return: the model
    :rtype: COPASI.CDataModel
    """
    global __current_model
    global __model_list
    __current_model = model
    if model not in __model_list:
        __model_list.append(model)

    if __current_model is not None:
        logging.debug(overview(__current_model))
    return __current_model


def get_current_model():
    """Returns the current model.

    This function returns the current model. That is the model loaded / created last. If no model exists
    a new one will be created first.

    :return: the current model
    :rtype: COPASI.CDataModel
    """
    global __current_model
    if __current_model is None:
        logging.warning('There is no model, creating a new one')
        new_model()
    return __current_model


def create_datamodel():
    """creates a new data model

    :return: new data model
    :rtype: COPASI.CDataModel
    """
    try:
        data_model = COPASI.CRootContainer.addDatamodel()
    except NameError:
        traceback.print_exc()
        data_model = COPASI.CCopasiRootContainer.addDatamodel()
    return data_model


def remove_datamodel(model):
    """Removes the model from the internal list of loaded models.

    The model is removed from the list of models, and current model,
    before it is freed

    :param model: the model to be removed

    :type model:COPASI.CDataModel
    :return: None
    """
    global __model_list
    global __current_model
    if model in __model_list:
        __model_list.remove(model)
    if __current_model is model:
        __current_model = None
    COPASI.CRootContainer.removeDatamodel(model)
    return None


def get_num_loaded_models():
    """
    :return: the number of loaded models
    """
    return COPASI.CRootContainer.getDatamodelList().size()


def new_model(**kwargs):
    """Creates a new model and sets it as current.
    
    :param kwargs: optional arguments

    - `name` (str): the name for the new model

    - `quantity_unit` (str): the unit to use for species

    - `time_unit` (str): the time unit to use

    - `volume_unit` (str): the unit to use for 3D compartments

    - `area_unit` (str): the unit to use for 2D compartments

    - `length_unit` (str): the unit to use for 1D compartments

    - | `notes`: sets notes for the model (either plain text, or valid xhtml)

    :return: the new model
    :rtype: COPASI.CDataModel
    """
    dm = create_datamodel()
    dm.newModel()

    model = dm.getModel()
    assert (isinstance(model, COPASI.CModel))

    if 'name' in kwargs:
        model.setObjectName(kwargs['name'])

    if 'quantity_unit' in kwargs:
        model.setQuantityUnit(kwargs['quantity_unit'])

    if 'area_unit' in kwargs:
        model.setAreaUnit(kwargs['area_unit'])

    if 'length_unit' in kwargs:
        model.setLengthUnit(kwargs['length_unit'])

    if 'volume_unit' in kwargs:
        model.setVolumeUnit(kwargs['volume_unit'])

    if 'time_unit' in kwargs:
        model.setTimeUnit(kwargs['time_unit'])

    if 'notes' in kwargs:
        model.setNotes(kwargs['notes'])

    return set_current_model(dm)


def load_model_from_string(content):
    """Loads either COPASI model / SBML model from the raw string given.

    :param content: the copasi / sbml model serialized as string
    :type content: str

    :return: the loaded model
    :rtype: COPASI.CDataModel
    """
    model = create_datamodel()
    if '<COPASI ' in content and model.loadModelFromString(content, os.getcwd()):
        return set_current_model(model)

    if '<sbml ' in content and model.importSBMLFromString(content):
        return set_current_model(model)

    return remove_datamodel(model)


def load_model_from_url(url):
    """Loads either COPASI model / SBML model from the url.

    :param url: url to a copasi / sbml model
    :type url: str

    :return: the loaded model
    :rtype: COPASI.CDataModel
    """
    if _use_urllib2:
        content = urllib2.urlopen(url).read()
    else:
        data = urllib.request.urlopen(url).read()
        content = data.decode("utf8")

    return load_model_from_string(content)


def load_model(location):
    """Loads the model and sets it as current

    :param location: either a filename, url or raw string of a COPASI / SBML model
    :type location: str
    :return: the loaded model
    :rtype: COPASI.CDataModel
    """
    model = create_datamodel()

    if os.path.isfile(location):
        if model.importSBML(location):
            return set_current_model(model)
        if model.loadModel(location):
            return set_current_model(model)

    remove_datamodel(model)

    model_from_string = load_model_from_string(location)
    if model_from_string is not None:
        return model_from_string

    if 'https://' in location or 'http://' in location:
        return load_model_from_url(location)

    return None


def load_biomodel(model_id):
    """Loads a model from the BioModels Database.

    :param model_id: either an integer of the biomodels id, or a valid biomodels id
    :type model_id: Union[int,str]
    :return:
    :rtype: COPASI.CDataModel
    """

    from . import biomodels

    return load_model_from_string(biomodels.get_content_for_model(model_id))


def get_examples(selector=''):
    """Returns the filenames of examples bundled with this version.

    A number of example models are included with the distribution. This method returns the filenames
    to those examples. Filtered by the argument.

    :param selector: a filter expression to be used, only files matching `*{selector}.[cps|xml]` will be returned
    :type selector: str
    :return: the list of examples matching
    :rtype: [str]
    """
    dir_name = os.path.dirname(__file__)
    types = ('*.xml', '*.cps')
    files = []
    for ext in types:
        files.extend(glob.glob(os.path.join(dir_name, 'data', '*{0}{1}'.format(selector, ext))))

    return files


def load_example(selector):
    """Loads the example matching the selector.

    :param selector: the filter expression to use for the examples see :func:`get_examples`
    :type selector: str
    :return: the loaded model, or None, if none matched
    :rtype: COPASI.CDataModel or None
    """
    files = get_examples(selector)

    if not files:
        return None

    return load_model(files[0])


def overview(model=None):
    """Returns a basic representation of the model.

    :param model: the model to get the overview for
    :type model: COPASI.CDataModel or None

    :return: a string, consisting of name, # compartments, # species, # parameters, # reaction
    :rtype: str
    """
    if model is None:
        model = get_current_model()

    c_model = model.getModel()

    assert (isinstance(c_model, COPASI.CModel))
    result = "Name:           {0}\n".format(c_model.getObjectName())
    if c_model.getNumCompartments() > 1:
        result += "# Compartments: {0}\n".format(c_model.getNumCompartments())
    if c_model.getNumMetabs() > 0:
        result += "# Species:      {0}".format(c_model.getNumMetabs())
    if c_model.getNumModelValues() > 0:
        result += "# Parameters:   {0}".format(c_model.getNumModelValues())
    if c_model.getNumReactions() > 0:
        result += "# Reactions:    {0}".format(c_model.getNumReactions())
    if c_model.getNumEvents() > 0:
        result += "# Events:       {0}".format(c_model.getNumEvents())

    return result


def print_model(model=None):
    """Prints the model overview.

    See also :func:`overview`

    :param model: the model
    :type model: COPASI.CDataModel

    :return: None
    """
    print(overview(model))


def save_model(filename, **kwargs):
    """Saves the model to the given filename.

    :param filename: the file to be written
    :type filename: str

    :param kwargs: optional arguments:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    - `type` (str): `copasi` to write COPASI files, `sbml` to write SBML files (defaults to `copasi`)

    - `overwrite` (bool): whether the file should be overwritten if present (defaults to True)

    - `level` (int): SBML level to export

    - `version` (int): SBML version to export

    - `export_copasi_miriam` (bool): whether to export copasi miriam annotations

    - `export_incomplete` (bool): whether to export incomplete SBML model

    :return: None
    """
    # type: (str, **kwargs) -> None
    model = kwargs.get('model', get_current_model())
    assert (isinstance(model, COPASI.CDataModel))
    file_type = kwargs.get('type', 'copasi').lower()
    overwrite = kwargs.get('overwrite', True)
    level = kwargs.get('level', 2)
    version = kwargs.get('version', 4)
    export_copasi_miriam = kwargs.get('export_copasi_miriam', True)
    export_incomplete = kwargs.get('export_incomplete', True)

    exporters = {
        'sbml': lambda filename: model.exportSBML(filename, overwrite, sbmlLevel=level, sbmlVersion=version,
                                                  exportIncomplete=export_incomplete,
                                                  exportCOPASIMIRIAM=export_copasi_miriam),
        'copasi': lambda filename: model.saveModel(filename, overwrite),
    }

    if file_type in exporters:
        try:
            if not exporters[file_type](filename):
                logging.warning("Saving the file as {0} failed with: \n{1}".
                                format(os.path.basename(filename), COPASI.CCopasiMessage.getAllMessageText()))
        except COPASI.CCopasiException:
            logging.error("Couldn't save the file as {0}".format(os.path.basename(filename)))


def save_model_to_string(**kwargs):
    """Saves the current model to string

    :param kwargs: optional arguments:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    :return: the copasi model as string
    :rtype: str

    """
    model = kwargs.get('model', get_current_model())
    assert (isinstance(model, COPASI.CDataModel))
    return model.saveModelToString()


def save_model_and_data(filename, **kwargs):
    """Saves the model to the give filename, along with all experimental data files.

    :param filename: the filename of the COPASI file to write
    :param filename: str
    :param kwargs: optional arguments:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    - | `delete_data_on_exit` (bool): a flag indicating whether the files should be deleted at the end of the
      | python session (defaults to False)

    :return: None
    """

    try:
        from . import task_parameterestimation
    except ImportError:
        import task_parameterestimation

    model = kwargs.get('model', get_current_model())
    assert (isinstance(model, COPASI.CDataModel))

    has_validations = task_parameterestimation.num_validations_files(**kwargs)
    has_experiments = task_parameterestimation.num_experiment_files(**kwargs)

    delete_data_on_exit = kwargs.get('delete_data_on_exit', False)

    global __temp_files

    if has_experiments + has_validations > 0:

        data_dir = os.path.dirname(filename)
        if not data_dir:
            data_dir = '.'

        old_names = {}
        new_names = {}

        model = kwargs.get('model', get_current_model())
        assert (isinstance(model, COPASI.CDataModel))

        task = model.getTask("Parameter Estimation")
        assert (isinstance(task, COPASI.CFitTask))

        problem = task.getProblem()
        assert (isinstance(problem, COPASI.CFitProblem))

        exp_set = problem.getExperimentSet()
        assert (isinstance(exp_set, COPASI.CExperimentSet))

        # copy experiment files
        for old_name in exp_set.getFileNames():
            new_name = os.path.join(data_dir, os.path.basename(old_name))
            if new_name == old_name:
                # same file skipping
                continue

            if not os.path.exists(old_name):
                logging.warning("Experimental data file {0} does not exist, the resulting COPASI file cannot"
                                " be used for Parameter Estimation".format(old_name))
                continue

            if os.path.exists(new_name):
                logging.warning("Experimental data file {0} does already exist, and will not be overwritten.".format(new_name))
            else:
                shutil.copyfile(old_name, new_name)
            old_names[old_name] = new_name
            new_names[new_name] = old_name
            if delete_data_on_exit:
                __temp_files.append(new_name)

        # rename experiments
        for i in range(problem.getExperimentSet().size()):
            experiment = problem.getExperimentSet().getExperiment(i)
            assert (isinstance(experiment, COPASI.CExperiment))

            new_name = old_names.get(experiment.getFileNameOnly(), '')
            if not new_name:
                continue

            experiment.setFileName(new_name)

        # export model to string
        model.saveModel(filename, True)

        # rename experiments
        for i in range(problem.getExperimentSet().size()):
            experiment = problem.getExperimentSet().getExperiment(i)
            assert (isinstance(experiment, COPASI.CExperiment))

            old_name = new_names.get(experiment.getFileNameOnly(), '')
            if not old_name:
                continue

            experiment.setFileName(old_name)

    else:
        model.saveModel(filename, True)


def open_copasi(**kwargs):
    """Saves the model as COPASI file and opens it in COPASI.

    The file will be written to a temporary file, and then it will be executed, so that the
    application registered to open it will start.

    :param kwargs: optional arguments:

    - | `model`: to specify the data model to be used (if not specified
      | the one from :func:`.get_current_model` will be taken)

    - | `filename` (str): the file name to write to, if not given a temp file will be created
      | that will be deleted at the end of the python session.

    :return: None
    """
    model = kwargs.get('model', get_current_model())
    assert (isinstance(model, COPASI.CDataModel))
    name = kwargs.get('filename', '')

    if not name:
        global __temp_dirs, __temp_files
        temp_name = tempfile.mkdtemp()
        __temp_dirs.append(temp_name)

        name = os.path.join(temp_name, 'model.cps')
        __temp_files.append(name)

        logging.info('using temp file: ' + name)
        logging.warning("Created a temporary file to open. This file will be deleted later, so safe your changes.")
        save_model_and_data(name, delete_data_on_exit=True, **kwargs)

    if platform.system() == 'Darwin':
        subprocess.call(('open', name))
    elif platform.system() == 'Windows': # Windows
        os.startfile(name)
    else:  # linux
        subprocess.call(('xdg-open', name))


@atexit.register
def __cleanup():
    # clean up temp files
    global __temp_files
    for name in __temp_files:
        if os.path.exists(name):
            os.remove(name)

    global __temp_dirs
    for name in __temp_dirs:
        if os.path.exists(name):
            try:
                os.removedirs(name)
            except OSError:
                logging.warning("Couldn't remove temp dir: {0}".format(name))



"""BasiCO is a simplified interface to COPASI.

This module provides convenience functions to quickly get a model loaded and simulated:

Example:

    >>> from basico import *
    >>> load_biomodel(10)
    >>> run_time_course().plot()

"""

from .model_io import *
from .model_info import *

from .task_timecourse import *
from .task_parameterestimation import *
from .task_steadystate import *
from .task_scan import *
from .task_optimization import *

from .array_tools import *
from .compartment_array_tools import *

settings = {}

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import logging

import numpy as np
import petab
import petab.simulate
import basico
import pandas as pd
import copasi_petab_importer
import os


def create_simulation_df(measurement_df, simulation_results):
    """Creates a simulation data frame

    :param measurement_df: the measurement data frame of the petab problem
    :type measurement_df: pd.DataFrame
    :param simulation_results: the simulation results as obtained from basico.get_simulation_results()
    :type simulation_results: []
    :return: simulation data frame with simulation results in the same format as the measurement data frame
    :rtype: pd.DataFrame
    """
    sim = measurement_df.copy(True)
    sim = sim.rename(columns={"measurement": "simulation"})
    exp_names = basico.get_experiment_names()
    for i in range(len(exp_names)):
        cond_id = exp_names[i]
        s_df = simulation_results[1][i]
        _update_df_from_simulation(sim, s_df, cond_id)
    return sim


def _update_df_from_simulation(petab_df, basico_df, experiment_name):
    """ Updates a simulation data frame with simulation results from one experiment

    :param petab_df: the petab simulation data frame to update
    :type petab_df: pd.DataFrame
    :param basico_df: a basico data frame of simulation results, whose values should be written into `s_df`
    :type basico_df: pd.DataFrame
    :param experiment_name: the experiment (or condition_id) to update in `s_df`
    :type experiment_name: str
    :return: None
    """
    sim_col = petab_df.columns.get_loc('simulation')
    sim_times = basico_df.index.to_list()
    have_time = 'Time' in basico_df.reset_index().columns
    for obs in basico_df.columns.to_list():
        obs_id = obs
        if obs == 'Time':
            continue
        if 'Values[' in obs:
            obs_id = obs[len('Values['):-1]

        values = basico_df[obs].to_list()
        num_values = len(values)

        sub_set = petab_df.loc[
            (petab_df.observableId == obs_id) & (petab_df.simulationConditionId == experiment_name), 'simulation']
        num_expected = len(sub_set)
        if num_expected == 0:
            # this observable is not in the measurements, so we dont have values to update
            continue
        elif num_expected != num_values:
            # we only have certain times and will have to update them individually
            expected = petab_df.loc[
                (petab_df.observableId == obs_id) & (petab_df.simulationConditionId == experiment_name)
            ][['time', 'simulation']]
            for index, row in expected.iterrows():
                if row.time in sim_times and have_time:
                    basic_data = basico_df.loc[row.time, obs]
                    if basic_data.size > 1:
                        # if there are more rows, we still only can take the first one
                        basic_data = basic_data.iloc[0]
                    petab_df.iloc[index, sim_col] = basic_data
                if np.isinf(row.time):
                    # update from steady state:
                    petab_df.iloc[index, sim_col] = basico_df[obs]
            continue

        petab_df.loc[
            (petab_df.observableId == obs_id) & (petab_df.simulationConditionId == experiment_name), 'simulation'
        ] = values


def petab_llh(pp, sim):
    """Computes the llh based on the simulation data frame

    :param pp: petab problem
    :type pp: petab.Problem

    :param sim: simulation data frame
    :type sim: pd.DataFrame

    :return: the llh as calculated by petab

    """
    return petab.calculate_llh(pp.measurement_df, sim, pp.observable_df, pp.parameter_df)


def petab_chi2(pp, sim):
    """Computes the chi2 based on the simulation data frame

    :param pp: petab problem
    :type pp: petab.Problem

    :param sim: simulation data frame
    :type sim: pd.DataFrame

    :return: the chi2 as calculated by petab

    """
    return petab.calculate_chi2(pp.measurement_df, sim, pp.observable_df, pp.parameter_df)


def petab_residuals(pp, sim):
    """Computes the residuals based on the simulation data frame

    :param pp: petab problem
    :type pp: petab.Problem

    :param sim: simulation data frame
    :type sim: pd.DataFrame

    :return: the residuals as calculated by petab

    """
    return petab.calculate_residuals(pp.measurement_df, sim, pp.observable_df, pp.parameter_df)


def write_problem_to(problem, output_dir, model_name):
    """Writes the given petab problem to the output directory with the given model_name

    :param problem: the petab problem
    :type problem: petab.Problem
    :param output_dir: the output directory where the files should be written to
    :type output_dir: str
    :param model_name: the common model name that all files will have
    :type model_name: str
    :return: dictionary of all the files created
    :rtype: {}
    """
    yaml_file = os.path.join(output_dir, 'problem_{0}.yaml'.format(model_name))
    model_file = os.path.join(output_dir, 'model_{0}.xml'.format(model_name))
    cond_file = os.path.join(output_dir, 'experimentalCondition_{0}.tsv'.format(model_name))
    data_file = os.path.join(output_dir, 'measurementData_{0}.tsv'.format(model_name))
    param_file = os.path.join(output_dir, 'parameters_{0}.tsv'.format(model_name))
    obs_file = os.path.join(output_dir, 'observables_{0}.tsv'.format(model_name))
    files = {
        'model': model_file,
        'conditions': cond_file,
        'measurements': data_file,
        'parameters': param_file,
        'observables': obs_file,
        'problem': yaml_file,
    }
    # write to files
    problem.to_files(
        sbml_file=model_file,
        condition_file=cond_file,
        measurement_file=data_file,
        parameter_file=param_file,
        observable_file=obs_file,
        yaml_file=yaml_file
    )
    return files


def load_petab(yaml_file, output_dir, out_name=None):
    """Loads the petab file as specified in the yaml file

    This file will be converted using the petab_copasi_importer, and serialized to the specified
    output directory.

    :param yaml_file: the yaml file of the petab description
    :type yaml_file: str
    :param output_dir: the output directory to write the files in
    :type output_dir: str
    :param out_name: optional base name of the copasi file. if it is not specified ith will be
        the basename of the yaml file
    :type out_name: str or None

    :return: the full path to the cps file loded
    """
    # convert to COPASI file
    converter = copasi_petab_importer.PEtabConverter.from_yaml(yaml_file, out_dir=output_dir, out_name=out_name)
    converter.convert()

    # load into basico
    cps_file = os.path.join(output_dir, converter.out_name + '.cps')
    basico.load_model(cps_file)
    return cps_file


class PetabSimulator(petab.simulate.Simulator):
    """ Implementation of a COPASI simulator with the petab simulation interface

        Examples:

        To instantiate the simulatior, just give it the problem and the settings you would like
        to use

        >>> sim = PetabSimulator(pp, settings= {'method' : {'name': PE.EVOLUTIONARY_STRATEGY_SRES}})
        >>> sim.simulate()

    """

    def __init__(self, petab_problem, working_dir=None, settings=None, name='', **kwargs):
        """ Inititializes the simulator

        :param petab_problem: A PEtab problem.
        :type petab_problem: petab.Problem

        :param working_dir: optional working dir (temp directory will be taken if it does not exist
        :type working_dir: Optional[Union[pathlib.Path, str]]

        :param settings: the settings to use for the parameter estimation (by default the converter will
               setup 'Current Solution Statistic', i.e: no parameter estimation will actually be carried out
               and just the values are returned.
        :type settings: {} or None

        :param name: an optional name that will be passed as model specific suffix to use while creating temp
               files (since the petab_problem will be serialized to files in order to be imported)s
        :type name: str

        :param kwargs: will be passed to base class. If it contains an argument:

            * 'cps_file' (str): this COPASI file will be loaded,
        """
        super().__init__(petab_problem, working_dir, **kwargs)
        self.settings = settings
        self.name = name
        self.cps_file = None
        self.load_model(**kwargs)

    def load_model(self, **kwargs):
        """Loads the model into basico

        If the `cps_file` is in the arguments, then this will be loaded. otherwise this instances
        petab_problem, will be written to files, converted to COPASI and then loaded (at which point
        the cps_file generated will be set).

        :param kwargs:

            - 'cps_file' (str): if present this file will be loaded rather than the petab problem

        :return: None
        """
        if 'cps_file' in kwargs:
            self.cps_file = kwargs['cps_file']
            basico.load_model(self.cps_file)
        else:
            files = write_problem_to(self.petab_problem, self.working_dir, self.name)
            self.cps_file = load_petab(files['problem'], self.working_dir)

    def evaluate(self):
        if self.settings is not None:
            basico.set_task_settings('Parameter Estimation', self.settings)
        return basico.run_parameter_estimation()

    def simulate_without_noise(self):
        """Simulate the PEtab problem.

        This is an abstract method that should be implemented with a simulation
        package. Examples of this are referenced in the class docstring.

        :return:
            Simulated data, as a PEtab measurements table, which should be
            equivalent to replacing all values in the `petab.C.MEASUREMENT`
            column of the measurements table (of the PEtab problem supplied to
            the `__init__` method), with simulated values.

        :rtype: pandas.DataFrame

        """

        simulation_results = basico.get_simulation_results(values_only=True, solution=self.evaluate())

        return create_simulation_df(self.petab_problem.measurement_df, simulation_results)

    def __getstate__(self):
        """Return state for pickling"""
        state = self.__dict__.copy()

        # maybe save cps file to omex? and add to state?

        return state

    def __setstate__(self, state):
        """Set state after unpickling"""
        self.__dict__.update(state)
        # initialize from 'cps_file' in state, otherwise we'd initialize from the settings
        self.load_model()



import logging
from math import log, isnan
import os
import tempfile

from . import core
import basico
import petab_select


def copasi_aic():
    """Calculates the AIC based on the last parameter estimation run

        SSR = weighted sum of squares
        N = number of data points
        K = number of parameters being estimated (note that this is one more than the number of parameters
        in COPASI, since the SSR is also a parameter being estimated; so this is basically K = parameters + 1)
        AIC = Akaike's information criterion
        AICc= corrected Akaike's information criterion (for low N)

        AIC = N * log(SSR/N) + 2*K

        AICc = N * log(SSR/N) + 2*K*(K+1)/(N-K-1)

        :return: the aic based on above calculation
        :rtype: float
    """
    prob = basico.get_current_model().getTask("Parameter Estimation").getProblem()
    N = prob.getExperimentSet().getValidValueCount()
    K = prob.getOptItemSize() + 1
    SSR = prob.getSolutionValue()
    if isnan(SSR) or N == 0:
        return float("inf")
    return N * log(SSR / N) + 2 * K


def default_evaluation():
    """Default evaluation of current model:

    * runs particle swarm for 600 iterations followed by
    * levelberg marquardt


    :return: found parameter values
    """
    basico.run_parameter_estimation(method='Particle Swarm', update_model=True,
                                    settings={'method': {'Iteration Limit': 600}})
    sol = basico.run_parameter_estimation(method='Levenberg - Marquardt', update_model=True)
    return sol


def evaluate_model(test_model, evaluation=default_evaluation, temp_dir=None, delete_temp_files=True,
                   sim_dfs=None, sol_dfs=None, temp_files=None):
    """evaluates the given test model and updates it with the calculated metrics and estimated parameters

    :param test_model: the model to test
    :type test_model: petab_select.Model
    :param evaluation: optional function to evaluate the test model with. defaults to func:`.default_evaluation`
    :type evaluation: () -> pandasDataFrame
    :param temp_dir: optional temp directory to store the files in (otherwise the os temp dir will be used)
    :type temp_dir: str or None
    :param delete_temp_files: boolean indicating whether temp files should be deleted
    :type delete_temp_files: bool
    :param sim_dfs: optional array, in which simulation data frames will be returned
    :type sim_dfs: [] or None
    :param sol_dfs: optional array in which found parameters will be returned
    :type sol_dfs: [] or None
    :param temp_files: optional array that returns filenames of temp files created during the run
    :type temp_files: [] or None
    :return: COPASI objective value of the evaluation
    :rtype: float
    """
    # create petab problem
    pp = test_model.to_petab()[0]

    created_temp_dir = False
    if temp_dir is None:
        temp_dir = tempfile.mkdtemp()
        created_temp_dir = True

    model_id = test_model.model_id
    files = core.write_problem_to(pp, temp_dir, model_id)

    # load into basico
    out_name = 'cps_{0}'.format(model_id)
    cps_file = os.path.join(temp_dir, out_name + '.cps')
    core.load_petab(files['problem'], temp_dir, out_name)

    files = files.values()
    files.append(cps_file)

    files = files + basico.get_experiment_filenames()

    # run parameter estimation
    sol = evaluation()
    if sol_dfs:
        sol_dfs.append(sol)

    simulation_results = basico.get_simulation_results(values_only=True, solution=sol)
    basico.prune_simulation_results(simulation_results)
    sim_df = basico.petab.create_simulation_df(pp.measurement_df, simulation_results)
    if sim_dfs:
        sim_dfs.append(sim_df)

    # compute metrics
    llh = basico.petab.petab_llh(pp, sim_df)
    aic = float(petab_select.calculate_aic(pp.parameter_df.estimate.sum(), -llh))
    aicc = float(petab_select.calculate_aicc(pp.parameter_df.estimate.sum(), -llh, len(pp.measurement_df), 0))
    bic = float(petab_select.calculate_bic(pp.parameter_df.estimate.sum(), -llh, len(pp.measurement_df), 0))
    aic_copasi = copasi_aic()
    task = basico.get_current_model().getTask("Parameter Estimation")
    prob = task.getProblem()
    obj = prob.getSolutionValue()
    logging.debug(
        'AIC: {0}, AICc: {1}, BIC: {2}, Copasi AIC: {3}, LLH: {4}, OBJ: {5}, RMS: {6}'
        .format(aic, aicc, bic, aic_copasi, llh, obj, prob.getRMS())
    )
    logging.debug(sol)
    test_model.set_criterion('AIC', aic)
    test_model.set_criterion('AICc', aicc)
    test_model.set_criterion('BIC', bic)

    # update estimated parameters
    for param_id in test_model.parameters:
        value = test_model.parameters[param_id]
        if not isnan(value):  # we only want to include what we estimated
            continue
        name = 'Values[{0}]'.format(param_id)
        if name in sol.index:
            test_model.estimated_parameters[param_id] = sol.loc[name].sol

    # write result for testing
    result_file = os.path.join(temp_dir, 'result_{0}.yaml'.format(model_id))
    files.append(result_file)
    test_model.to_yaml(result_file)

    # delete temp files if needed
    if delete_temp_files:
        for file in files:
            os.remove(file)
        if created_temp_dir:
            # since we created the temp dir, lets get rid of it
            os.rmdir(temp_dir)
    elif temp_files is not None:
        # add temp files to the list of temp files:
        temp_files = temp_files + files

    return obj


def evaluate_models(test_models, evaluation=default_evaluation, temp_dir=None, delete_temp_files=True,
                    sim_dfs=None, sol_dfs=None, temp_files=None):
    """Evaluates all temp models iteratively

    :param test_models: the models to evaluate
    :param evaluation: optional function to evaluate the test model with. defaults to func:`.default_evaluation`
    :type evaluation: () -> pandasDataFrame
    :param temp_dir: optional temp directory to store the files in (otherwise the os temp dir will be used)
    :type temp_dir: str or None
    :param delete_temp_files: boolean indicating whether temp files should be deleted
    :type delete_temp_files: bool
    :param sim_dfs: optional array, in which simulation data frames will be returned
    :type sim_dfs: [] or None
    :param sol_dfs: optional array in which found parameters will be returned
    :type sol_dfs: [] or None
    :param temp_files: optional array that returns filenames of temp files created during the run
    :type temp_files: [] or None
    :return:
    """

    for test_model in test_models:
        try:
            evaluate_model(test_model, evaluation, temp_dir, delete_temp_files, sim_dfs, sol_dfs, temp_files)
        except:
            logging.exception("couldn't evaluate " + test_model.model_id)


def evaluate_problem(selection_problem, candidate_space=None, evaluation=default_evaluation, temp_dir=None,
                     delete_temp_files=True, sim_dfs=None, sol_dfs=None, temp_files=None):
    """Evaluates the given selection problem with the specified candidate space returning the best model found

    :param selection_problem: the selection problem
    :type selection_problem: petab_select.Problem
    :param candidate_space: optional the candidate space to use (otherwise the one from the problem method will be used)
    :type candidate_space: petab_select.CandidateSpace or None
    :param evaluation: optional function to evaluate the test model with. defaults to func:`.default_evaluation`
    :type evaluation: () -> pandasDataFrame
    :param temp_dir: optional temp directory to store the files in (otherwise the os temp dir will be used)
    :type temp_dir: str or None
    :param delete_temp_files: boolean indicating whether temp files should be deleted
    :type delete_temp_files: bool
    :param sim_dfs: optional array, in which simulation data frames will be returned
    :type sim_dfs: [] or None
    :param sol_dfs: optional array in which found parameters will be returned
    :type sol_dfs: [] or None
    :param temp_files: optional array that returns filenames of temp files created during the run
    :type temp_files: [] or None
    :return: the best model found
    :rtype: petab_select.Model
    """
    model_space = selection_problem.model_space

    if candidate_space is None:
        logging.info('initializing new candidate space with method: {0}'.format(selection_problem.method))
        if selection_problem.method == 'forward':
            candidate_space = petab_select.ForwardCandidateSpace(get_initial_virtual_model(selection_problem))
        elif selection_problem.method == 'brute_force':
            candidate_space = petab_select.BruteForceCandidateSpace()
        elif selection_problem.method == 'backward':
            candidate_space = petab_select.BackwardCandidateSpace(
                get_initial_virtual_model(selection_problem, initial_value=float('nan')))
        else:
            logging.error('unsupported candidate space {0} please initialize first and pass along'
                          .format(selection_problem.method))

    test_models = model_space.neighbors(candidate_space)

    if not test_models:
        logging.warning('no models to test, method: {0}'.format(selection_problem.method))
        return None

    chosen_model = None
    while test_models:
        basico.petab.evaluate_models(test_models, evaluation, temp_dir, delete_temp_files, sim_dfs, sol_dfs, temp_files)

        for model in test_models:
            logging.info('{0} = {1}'.format(model.model_id, model.criteria))

        chosen_model = selection_problem.get_best(test_models)
        if chosen_model is None:
            logging.warning('found no best model?')

        logging.debug('best model is {0}'.format(chosen_model.model_id))
        candidate_space.reset(chosen_model)
        test_models = model_space.neighbors(candidate_space)

    return chosen_model


def get_initial_virtual_model(selection_problem, initial_value=0.01):
    """Utility function returning an initial virtual model

    :param selection_problem: the selection problem
    :type selection_problem: petab_select.Problem
    :param initial_value: initial value for all parameters to use (defaults to 0.01)
    :type initial_value: float

    :return: the initial virtual model
    :rtype: petab_select.Model
    """
    model_space = selection_problem.model_space
    parameters = {
        k: initial_value
        for k in model_space.parameter_ids
    }

    initial_virtual_model = petab_select.Model(
        model_id='INITIAL_VIRTUAL_MODEL',
        petab_yaml=model_space.source_path,
        parameters=parameters,
    )

    return initial_virtual_model



from . import model_io
import COPASI
import logging
import re
import pandas
import numpy as np


def compartment_names(**kwargs):
    # type: (**kwargs) -> [str]
    dm = kwargs.get('model', model_io.get_current_model())
    assert (isinstance(dm, COPASI.CDataModel))

    assert (isinstance(dm, COPASI.CDataModel))
    model = dm.getModel()
    assert (isinstance(model, COPASI.CModel))

    compartments = model.getCompartments()
    assert (isinstance(compartments, COPASI.CompartmentVectorNS))
    num_compartments = compartments.size()

    names = []

    # an array model has to have at least one chain comp[1] -> comp[2]
    for i in range(num_compartments):
        comp = compartments.get(i)
        assert (isinstance(comp, COPASI.CCompartment))
        name = comp.getObjectName()
        names.append(name)

    return names


def remove_template(**kwargs):
    """Removes the template model from the array model"""
    names = compartment_names(**kwargs)
    dm = kwargs.get('model', model_io.get_current_model())
    assert (isinstance(dm, COPASI.CDataModel))
    model = dm.getModel()
    assert (isinstance(model, COPASI.CModel))
    exp = re.compile(r'([-a-zA-Z_0-9 ]+)\\[(\d+)\\]')

    compartments = model.getCompartments()
    assert (isinstance(compartments, COPASI.CompartmentVectorNS))

    for name in names:
        if not exp.match(name):
            comp = model.getCompartments().getByName(name)
            if comp is None:
                logging.warning("Could not find comp {0}".format(name))
            else:
                logging.warning("Removing {0} recursively".format(name))
                model.removeCompartment(comp.getKey(), True)
                model.compileIfNecessary()


def remove_arrays(**kwargs):
    """Removes arrays  from the model"""
    names = compartment_names(**kwargs)
    dm = kwargs.get('model', model_io.get_current_model())
    assert (isinstance(dm, COPASI.CDataModel))
    model = dm.getModel()
    assert (isinstance(model, COPASI.CModel))
    exp = re.compile(r'([-a-zA-Z_0-9 ]+)\\[(\d+)\\]')

    for name in names:
        if exp.match(name):
            comp = model.getCompartments().getByName(name)
            if comp is None:
                logging.warning("Could not find comp {0}".format(name))
            else:
                logging.warning("Removing {0} recursively".format(name))
                model.removeCompartment(comp.getKey(), True)
                model.compileIfNecessary()


def plot_arrays_1d(df, species, **kwargs):
    # type:(pandas.DataFrame, str, **kwargs) -> pandas.DataFrame
    dm = kwargs.get('model', model_io.get_current_model())
    assert (isinstance(dm, COPASI.CDataModel))
    names = compartment_names(**kwargs)

    candidates = []
    for name in df.columns:
        index = name.find('{')
        if index < 0:
            continue
        name = name[0:index]

        if species not in name:
            continue

        if name not in candidates:
            candidates.append(name)

    cols = []
    for compartment in names:
        for species in candidates:
            index = species + '{' + compartment + '}'
            if index not in df:
                continue
            cols.append(index)

    rows = []
    row_index = kwargs.get('index', df.index)
    num_rows = len(row_index)
    num_cols = len(cols)

    data = np.empty((num_rows, num_cols))

    for row in range(num_rows):
        rows.append(row_index[row])
        for col in range(num_cols):
            col_name = cols[col]
            data[row, col] = df[col_name][row_index[row]]

    return pandas.DataFrame(data, index=rows, columns=cols)
